from typing import List, Tuple

from vowpalwabbit import pyvw
from movado.mab_handler import MabHandler
import numpy as np


class MabHandlerCB(MabHandler):
    def __init__(self, debug: bool = False, epsilon: float = 0.2):
        super().__init__(debug)
        self._mab = pyvw.vw("--cb_explore 2 " + "--cover " + str(3))

    def predict(self, context: List[float]) -> int:
        context_str: str = "| "
        for feature in context:
            context_str += str(feature) + " "
        context_str.strip()
        probability_distribution: Tuple[float, float] = self._mab.predict(context_str)
        self._last_predict_probability = np.max(probability_distribution)
        return np.argmax(probability_distribution) + 1



from typing import List
from pathlib import Path
from abc import ABC, abstractmethod
import numpy as np


class MabHandler(ABC):
    def __init__(self, debug: bool = False):
        self._mab = None
        self._last_predict_probability: float = -1
        self._debug = debug
        self._sample_prefix = ""  # if changing this add trailing whitespace
        if debug:
            self.__costs: List[float] = []
            Path("movado_debug").mkdir(exist_ok=True)
            self.__mab_debug = "movado_debug/mab.csv"
            Path(self.__mab_debug).open("w").close()
            Path(self.__mab_debug).open("a").write("Mean Reward")

    @abstractmethod
    def predict(self, context: List[float]) -> float:
        pass

    def learn(
        self,
        action: float,
        cost: float,
        context: List[float],
        forced_predict_probability: int = None,
    ) -> None:
        sample: str = (
            self._sample_prefix
            + str(action)
            + ":"
            + str(cost)
            + ":"
            + (
                str(self._last_predict_probability)
                if not forced_predict_probability
                else str(forced_predict_probability)
            )
            + " | "
        )
        for feature in context:
            sample += str(feature) + " "
        self._mab.learn(sample)
        if self._debug:
            self.__costs.append(cost)
            Path(self.__mab_debug).open("a").write(str(np.mean(self.__costs)) + "\n")



from typing import List, Callable

from movado.controller import Controller
from movado.estimator import Estimator
from functools import wraps
from movado.distance_controller import DistanceController
from movado.hoeffding_adaptive_tree_model import HoeffdingAdaptiveTreeModel
from movado.chained_estimator import ChainedEstimator

# This imports are unused but populate the global symbol table for the globals() call
from movado.mab_controller import MabController
from movado.kernel_regression_model import KernelRegressionModel


def approximate(**kwargs):
    def approximate_decorator(func: Callable[[List[float]], List[float]]):
        controller: Controller
        estimator: Estimator
        is_first_call: bool = True

        @wraps(func)
        def wrapper(point):
            nonlocal is_first_call
            nonlocal controller
            nonlocal estimator
            if is_first_call:
                is_first_call = False
                selected_model = globals().get(str(kwargs.get("estimator")) + "Model")
                model = (
                    HoeffdingAdaptiveTreeModel if not selected_model else selected_model
                )
                # TODO add multi-output non-chained estimators
                estimator = ChainedEstimator(
                    model(
                        **{
                            k[k.find("_") + 1 :]: v
                            for k, v in kwargs.items()
                            if k[: k.find("_")] == "estimator"
                        }
                    )
                )
                selected_controller = globals().get(
                    str(kwargs.get("controller")) + "Controller"
                )
                controller = (
                    DistanceController
                    if not selected_controller
                    else selected_controller
                )
                controller = controller(
                    func,
                    estimator,
                    **{
                        k[k.find("_") + 1 :]: v
                        for k, v in kwargs.items()
                        if k[: k.find("_")] == "controller"
                    }
                )

            controller.compute_objective(
                point
            )  # we expect the first and only argument of the function to be the input point

        if kwargs.get("disabled"):
            return func
        return wrapper

    return approximate_decorator


class Movado:
    def __init__(self):
        self.models = [
            model.replace("Model", "") for model in globals().keys() if "Model" in model
        ]
        self.controllers = [
            controller.replace("Controller", "")
            for controller in globals().keys()
            if "Controller" in controller
        ]



from typing import List, Tuple

from vowpalwabbit import pyvw

from movado.mab_handler import MabHandler


class MabHandlerCATS(MabHandler):
    def __init__(
        self,
        mab_bandwidth: int = 1,
        epsilon: float = 0.2,
        debug=False,
    ):
        super(MabHandlerCATS, self).__init__(debug)
        self._sample_prefix = "ca "
        if mab_bandwidth < 0:
            raise Exception(
                "Invalid bandwidth value: "
                + str(mab_bandwidth)
                + " it must be greater or equal than 1"
            )
        if epsilon < 0 or epsilon > 1:
            raise Exception(
                "Invalid epsilon value: " + str(epsilon) + "it must be between 0 and 1"
            )
        mab_actions: int = 100
        self._mab: pyvw.vw = pyvw.vw(
            "--cats "
            + str(mab_actions)
            + "  --bandwidth "
            + str(mab_bandwidth)
            + " --min_value 0 --max_value 100 --chain_hash --coin --epsilon 0.2 -q ::"
        )

    def predict(self, context: List[float]) -> float:
        context_str: str = "| "
        for feature in context:
            context_str += str(feature) + " "
        context_str.strip()
        action: Tuple[float, float] = self._mab.predict(context_str)
        self._last_predict_probability = action[1]
        return action[0]



from abc import abstractmethod, ABC
from typing import Tuple, Optional, Callable, Any, List, Dict, Union
from sklearn.preprocessing import StandardScaler
import numpy as np
import time
from pathlib import Path

from movado.estimator import Estimator
from movado.mab_handler import MabHandler


class Controller(ABC):
    @abstractmethod
    def __init__(
        self,
        exact_fitness: Callable[[List[float]], List[float]],
        estimator: Estimator,
        debug=False,
    ):
        self.__scaler = StandardScaler()
        self.__time_dataset: List[float] = []
        self.__error_dataset: List[float] = []
        self.__to_reshape = False
        self._exact: Callable[[List[float]], List[float]] = exact_fitness
        self._estimator: Estimator = estimator
        self._debug = debug
        if self._debug:
            Path("movado_debug").mkdir(exist_ok=True)
            self._controller_debug = "movado_debug/controller.csv"
            Path(self._controller_debug).open("w").close()

    @abstractmethod
    def compute_objective(self, point: List[int]) -> float:
        pass

    def _compute_exact(
        self,
        point: List[int],
        mab: Optional[Tuple[MabHandler, Union[int, float]]] = None,
        mab_forced_probability: Optional[int] = None,
    ) -> Tuple[List[float], float]:
        exact, exec_time = Controller.measure_execution_time(self._exact, point)
        self._estimator.train(point, exact)
        if mab:
            mab[0].learn(
                mab[1], self.get_time_error_z_score(exec_time, 0), point
            ), mab_forced_probability
        return exact, exec_time

    def _compute_estimated(
        self,
        point: List[int],
        mab: Optional[Tuple[MabHandler, Union[int, float]]] = None,
    ) -> Tuple[List[float], float]:
        estimation, exec_time = Controller.measure_execution_time(
            self._estimator.predict, point
        )
        if mab:
            mab[0].learn(
                mab[1],
                self.get_time_error_z_score(exec_time, self._estimator.get_error()),
                point,
            )
        return estimation, exec_time

    @abstractmethod
    def _write_debug(self, debug_info: Dict[str, Any]):
        pass

    def get_time_error_z_score(self, response_time: float, error: float) -> float:
        self.__time_dataset.append(response_time)
        self.__error_dataset.append(error)
        if len(self.__time_dataset) > 1:
            reshaped_time = np.reshape(self.__time_dataset, (-1, 1))
            reshaped_error = np.reshape(self.__error_dataset, (-1, 1))
        else:
            reshaped_time = np.reshape(self.__time_dataset, (1, -1))
            reshaped_error = np.reshape(self.__error_dataset, (1, -1))
        self.__scaler.fit(reshaped_time)
        time_z_score: float = self.__scaler.transform(reshaped_time)[-1][0]
        self.__scaler.fit(reshaped_error)
        error_z_score: float = self.__scaler.transform(reshaped_error)[-1][0]
        return -time_z_score - error_z_score

    @staticmethod
    def measure_execution_time(
        func: Callable[..., Any], *args, **kwargs
    ) -> Tuple[Any, float]:
        start = time.time()
        out = func(*args, **kwargs)
        execution_time = time.time() - start
        return out, execution_time
